# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides additional stream classes."""

import asyncio
import logging


__all__ = ("StreamProtocol",)
"""Tuple all the things."""


logger = logging.getLogger("blackhole.streams")


class StreamProtocol(asyncio.streams.FlowControlMixin, asyncio.Protocol):
    """Helper class to adapt between Protocol and StreamReader."""

    def __init__(self, *, loop=None, **kwargs):
        """Stream protocol."""
        super().__init__(loop=loop)
        self.transport = None
        self.writer = None
        self.reader = asyncio.StreamReader(loop=loop)

    def is_connected(self):
        """Client is connected."""
        return self.transport is not None

    def connection_made(self, transport):
        """Client connection made callback."""
        self.transport = transport
        self.reader.set_transport(transport)
        self.writer = asyncio.StreamWriter(
            transport,
            self,
            self.reader,
            self._loop,
        )

    def connection_lost(self, exc):
        """Client connection lost callback."""
        self.transport = self.writer = None
        self.reader._transport = None

        logger.debug(exc)
        if exc is None:
            self.reader.feed_eof()
        else:
            self.reader.set_exception(exc)

        super().connection_lost(exc)

    def data_received(self, data):
        """Client data received."""
        self.reader.feed_data(data)

    def eof_received(self):
        """Client EOF received."""
        self.reader.feed_eof()



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Configuration structure and functionality for testing config validity."""


import argparse
import getpass
import grp
import inspect
import logging
import multiprocessing
import os
import pathlib
import pwd
import socket
import tempfile

from .exceptions import ConfigException
from .utils import Singleton, get_version, mailname


__all__ = ("parse_cmd_args", "warn_options", "config_test", "Config")
"""Tuple all the things."""


def parse_cmd_args(args):
    """
    Parse arguments from the command line.

    https://kura.gg/blackhole/configuration.html#command-line-options

    :param list args: Command line arguments.
    :returns: Parsed command line arguments.
    :rtype: :py:class:`argparse.Namespace`
    """
    ls_help = (
        "Disable ssl.OP_SINGLE_DH_USE and ssl.OP_SINGLE_ECDH_USE. "
        "Reduces CPU overhead at the expense of security -- Don't "
        "use this option unless you really need to."
    )

    description = (
        "Blackhole is an MTA (mail transfer agent) that "
        "(figuratively) pipes all mail to /dev/null. Blackhole is "
        "built on top of asyncio and utilises async def and await "
        "statements available in Python 3.5 and above."
    )

    epilog = (
        "An explanation of all command line and all configuration "
        "options is provided here -- "
        "https://kura.gg/blackhole/configuration.html"
    )

    parser = argparse.ArgumentParser(
        "blackhole",
        description=description,
        epilog=epilog,
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=get_version(),
    )
    parser.add_argument(
        "-c",
        "--conf",
        type=str,
        dest="config_file",
        metavar="FILE",
        help="override the default configuration options",
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-t",
        "--test",
        dest="test",
        action="store_true",
        help="perform a configuration test",
    )
    group.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        help="enable debugging mode",
    )
    group.add_argument(
        "-b",
        "--background",
        dest="background",
        action="store_true",
        help="run in the background",
    )
    group.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="store_true",
        help="Suppress warnings",
    )
    parser.add_argument(
        "-ls",
        "--less-secure",
        dest="less_secure",
        action="store_true",
        help=ls_help,
    )
    return parser.parse_args(args)


def warn_options(config):
    """
    Warn the user when using certain options.

    :param Config config: The configuration.
    """
    logger = logging.getLogger("blackhole.warnings")
    if config.args.less_secure:
        logger.warning(
            "Using -ls or --less-secure reduces security on "
            "SSL/TLS connections.",
        )
    if not config.tls_dhparams and len(config.tls_listen) > 0:
        logger.warning(
            "TLS is enabled but no Diffie Hellman ephemeral "
            "parameters file was provided.",
        )
    _compare_uid_and_gid(config)


def config_test(args):
    """
    Test the validity of the configuration file content.

    :param argparse.Namespace args: Parsed arguments.
    :raises SystemExit: Exit code :py:obj:`os.EX_OK` when config is valid
                        or :py:obj:`os.EX_USAGE` when config is invalid.

    .. note::

       Problems with the configuration will be written to the console using
       the :py:mod:`logging` module.
    """
    logger = logging.getLogger("blackhole.config.test")
    logger.setLevel(logging.INFO)
    try:
        conf = Config(args.config_file).load().test()
        conf.args = args
    except ConfigException:
        logger.critical("Config error")
        raise SystemExit(os.EX_USAGE)
    logger.info(f"blackhole: {args.config_file} syntax is OK.")
    logger.info(f"blackhole: {args.config_file} test was successful.")
    warn_options(conf)
    raise SystemExit(os.EX_OK)


def _compare_uid_and_gid(config):
    """
    Compare the current user and group and conf settings.

    :param Config config: The configuration.
    """
    logger = logging.getLogger("blackhole.warnings")
    uid, gid = os.getuid(), os.getgid()
    user, group = config.user, config.group
    if (uid == 0 and gid == 0) and (user == "root" and group == "root"):
        logger.warning(
            "It is unsafe to run Blackhole as root without setting "
            "a user and group for privilege separation.",
        )


class Config(metaclass=Singleton):
    """
    Configuration module.

    Default values are provided as well as self-test functionality
    to sanity check configuration.

    https://kura.gg/blackhole/configuration.html#configuration-options
    """

    args = None
    """Arguments parsed from the command line."""

    config_file = None
    """A file containing configuration values."""

    _workers = 1
    _listen = []
    _tls_listen = []
    _user = None
    _group = None
    _timeout = 60
    _tls_key = None
    _tls_cert = None
    _tls_dhparams = None
    _pidfile = os.path.join(tempfile.gettempdir(), "blackhole.pid")
    _delay = None
    _mode = "accept"
    _max_message_size = 512000
    _dynamic_switch = None

    def __init__(self, config_file=None):
        """
        Initialise the configuration.

        :param str config_file: The configuration file path. Default: ``None``
        """
        if config_file:
            self.config_file = config_file
        self.user = getpass.getuser()
        self.group = grp.getgrgid(os.getgid()).gr_name
        # this has to be cached here due to the socket.getfqdn call failing
        # in os.fork
        self.mailname = mailname()

    def load(self):
        """
        Load the configuration file and parse.

        :raises ConfigException: When the configuration is invalid.
        :returns: :class:`Config`.

        .. note::

           Spaces, single and double quotes will be stripped. Lines beginning
           # will be ignored. # comments in-line will be stripped out and
           ignored.

           i.e.

           # listen = :1025, :::1025   ->
           listen = :25, :::25  # IPv4 & IPv6   ->   listen = :25, :::25
        """
        if self.config_file is None:
            return self
        if not os.access(self.config_file, os.R_OK):
            msg = "Config file does not exist or is not readable."
            raise ConfigException(msg)
        with open(self.config_file, "r") as _conf_file:
            for line in _conf_file.readlines():
                line = line.strip()
                if line.startswith("#"):
                    continue
                if line.strip() == "":
                    continue
                if "#" in line:
                    line = line.split("#")[0]
                if line.count("=") >= 1:
                    key, value = line.split("=", 1)
                    key, value = key.strip(), value.strip()
                    self.validate_option(key)
                    value = value.replace('"', "").replace("'", "")
                    setattr(self, key, value)
                else:
                    self.validate_option(line)
        return self

    def validate_option(self, key):
        """
        Validate config option is actually... valid...

        https://kura.gg/blackhole/configuration.html#configuration-options

        :param str key: Configuration option.
        :raises ConfigException: When an invalid option is configured.
        """
        if key == "":
            return
        attributes = inspect.getmembers(
            self,
            lambda a: not (inspect.isroutine(a)),
        )
        attrs = [
            a[0][1:]
            for a in attributes
            if not (a[0].startswith("__") and a[0].endswith("__"))
            and a[0].startswith("_")
        ]
        if key not in attrs:
            _attrs = "', '".join(attrs[:-1])
            valid_attrs = f"'{_attrs}' and '{attrs[-1]}'"
            msg = (
                f"Invalid configuration option '{key}'.\n\nValid options "
                f"are: {valid_attrs}"
            )
            raise ConfigException(msg)

    @property
    def workers(self):
        """
        How many workers to spawn to handle incoming connections.

        https://kura.gg/blackhole/configuration.html#workers

        :returns: Number of workers. Default: ``1``
        :rtype: :py:obj:`int`

        .. note::

           Default value is 1.

           A supervisor process will always exist separately from the workers.
        """
        return int(self._workers) or 1

    @workers.setter
    def workers(self, workers):
        self._workers = workers

    @property
    def listen(self):
        """
        Address, port and socket family.

        https://kura.gg/blackhole/configuration.html#listen

        :returns: Listeners.
        :rtype: :py:obj:`list`

        .. note::

           Default IPv4:

               [('127.0.0.1', 25, socket.AF_INET,
                ('127.0.0.1', 587, socket.AF_INET), ]

           Default IPv6:

               [('127.0.0.1', 25, socket.AF_INET),
                ('127.0.0.1', 587, socket.AF_INET), ]
        """
        ipv4 = [
            ("127.0.0.1", 25, socket.AF_INET, {}),
            ("127.0.0.1", 587, socket.AF_INET, {}),
        ]
        ipv6 = [
            ("::", 25, socket.AF_INET6, {}),
            ("::", 587, socket.AF_INET6, {}),
        ]
        default = ipv4 + ipv6 if socket.has_ipv6 else ipv4
        return self._listen or default

    @listen.setter
    def listen(self, addrs):
        self._listen = self._listeners(addrs)

    @property
    def tls_listen(self):
        """
        Address and port and socket family for SSL/TLS connections.

        https://kura.gg/blackhole/configuration.html#tls-listen

        :returns: TLS listeners. Default: ``[]``
        :rtype: :py:obj:`list`
        """
        return self._tls_listen or []

    @tls_listen.setter
    def tls_listen(self, addrs):
        self._tls_listen = self._listeners(addrs)

    @property
    def user(self):
        """
        UNIX user.

        https://kura.gg/blackhole/configuration.html#user

        :returns: User name.
        :rtype: :py:obj:`str`

        .. note::

           Defaults to the current user.
        """
        return self._user

    @user.setter
    def user(self, user):
        self._user = user

    @property
    def group(self):
        """
        UNIX group.

        https://kura.gg/blackhole/configuration.html#group

        :returns: Group name.
        :rtype: :py:obj:`str`

        .. note::

           Defaults to the current group.
        """
        return self._group

    @group.setter
    def group(self, group):
        self._group = group

    @property
    def timeout(self):
        """
        Timeout in seconds.

        https://kura.gg/blackhole/configuration.html#timeout

        :returns: Timeout in seconds. Default: ``60``
        :rtype: :py:obj:`int`

        .. note::

           Defaults to 60 seconds.
           Cannot be more 180 seconds for security (denial of service).
        """
        return int(self._timeout)

    @timeout.setter
    def timeout(self, timeout):
        self._timeout = timeout

    @property
    def tls_key(self):
        """
        TLS key file.

        https://kura.gg/blackhole/configuration.html#tls-key

        :returns: Path to a TLS key file. Default: ``None``
        :rtype: :py:obj:`str`
        """
        return self._tls_key

    @tls_key.setter
    def tls_key(self, tls_key):
        if tls_key is not None:
            self._tls_key = tls_key

    @property
    def tls_cert(self):
        """
        TLS certificate file.

        https://kura.gg/blackhole/configuration.html#tls-cert

        :returns: Path to a TLS certificate. Default: ``None``
        :rtype: :py:obj:`str`
        """
        return self._tls_cert

    @tls_cert.setter
    def tls_cert(self, tls_cert):
        if tls_cert is not None:
            self._tls_cert = tls_cert

    @property
    def tls_dhparams(self):
        """
        Diffie Hellman ephemeral parameters.

        https://kura.gg/blackhole/configuration.html#tls-dhparams

        :returns: Path to a file containing dhparams. Default: ``None``
        :rtype: :py:obj:`str`
        """
        return self._tls_dhparams

    @tls_dhparams.setter
    def tls_dhparams(self, tls_dhparams):
        if tls_dhparams is not None:
            self._tls_dhparams = tls_dhparams

    @property
    def pidfile(self):
        """
        Path to store the pid.

        https://kura.gg/blackhole/configuration.html#pidfile

        :returns: Path to a pid file. Default: ``/tmp/blackhole.pid``.
        :rtype: :py:obj:`str`
        """
        return self._pidfile

    @pidfile.setter
    def pidfile(self, pidfile):
        if pidfile is not None:
            self._pidfile = pidfile

    @property
    def delay(self):
        """
        Delay in seconds.

        https://kura.gg/blackhole/configuration.html#delay

        :returns: Delay in seconds. Default: ``None``
        :rtype: :py:obj:`int` or :py:obj:`None`

        .. note::

           Defaults to :py:obj:`None`.
           Cannot be higher than 60 seconds for security (denial of service).
        """
        if self._delay is not None:
            return int(self._delay)
        return None

    @delay.setter
    def delay(self, delay):
        self._delay = delay

    @property
    def mode(self):
        """
        Mode with which to respond.

        https://kura.gg/blackhole/configuration.html#mode

        :returns: A response mode. Default: ``accept``.
        :rtype: :py:obj:`str`

        .. note::

           Defaults to 'accept'.
           Options: 'accept', 'bounce' and 'random'.
        """
        return self._mode

    @mode.setter
    def mode(self, mode):
        self._mode = mode.lower()

    @property
    def max_message_size(self):
        """
        Maximum size, in bytes, of a message.

        https://kura.gg/blackhole/configuration.html#max-message-size

        :returns: Maximum message size in bytes. Default: ``512000``.
        :rtype: :py:obj:`int`

        .. note::

           Default 512000 bytes (512 KB).
        """
        if self._max_message_size is not None:
            return int(self._max_message_size)
        return None

    @max_message_size.setter
    def max_message_size(self, size):
        self._max_message_size = size

    @property
    def dynamic_switch(self):
        """
        Enable or disable dynamic switches.

        https://kura.gg/blackhole/configuration.html#dynamic-switch

        :returns: Whether dynamic switches are enabled or not. Default:
                  ``True``.
        :rtype: :py:obj:`bool`

        .. note::

           Allowed values are :py:obj:`True` and :py:obj:`False`.
           Default: :py:obj:`True`
        """
        if self._dynamic_switch is None:
            return True
        return self._dynamic_switch

    @dynamic_switch.setter
    def dynamic_switch(self, switch):
        if switch.lower() == "false":
            self._dynamic_switch = False
        elif switch.lower() == "true":
            self._dynamic_switch = True
        else:
            msg = f"{switch} is not valid. Options are true or false."
            raise ConfigException(msg)

    def _convert_port(self, port):
        """
        Convert a port from the configuration files' string to an integer.

        :param str port: A port number.
        :type port: :py:obj:`str`
        :raises ConfigException: If an invalid port number if provided.
        :returns: A port number.
        :rtype: :py:obj:`int`
        """
        try:
            return int(port)
        except ValueError:
            msg = f"{port} is not a valid port number."
            raise ConfigException(msg)

    def _listeners(self, listeners):
        """
        Convert listeners lines from the configuration to usable values.

        :param str listeners: A list of addresses and ports, separated by
                              commas.
                              -- e.g. '127.0.0.1:25, 10.0.0.1:25, :25, :::25'
        :returns: List of addresses and sockets to listen on.
        :rtype: :py:obj:`list` or :py:obj:`None`
        """
        clisteners = []
        _listeners = listeners.split(",")
        if len(_listeners) == 0:
            return None
        for listener in _listeners:
            listener = listener.strip()
            parts = listener.split(" ")
            addr_port = parts[0]
            port = addr_port.split(":")[-1].strip()
            addr = addr_port.replace(f":{port}", "").strip()
            family = socket.AF_INET
            if ":" in addr:
                family = socket.AF_INET6
            flags = {}
            if len(parts) > 1:
                flags = self.create_flags(parts[1:])
            host = (addr, self._convert_port(port), family, flags)
            clisteners.append(host)
        return clisteners

    def flags_from_listener(self, addr, port):
        """
        Get a list of flags defined for the provided listener.

        Scope: ``listen``, ``tls_listen``.

        :param str addr: The listener host address.
        :param int port: The listener port.
        :returns: Flags defined for this socket. Default: ``{}``.
        :rtype: :py:obj:`dict`

        .. note::

           If multiple modes or delays are listed in a single listener, the
           last definition will be used:

           ``listen = :25 mode=bounce mode=random``  ->  ``mode=random``

           A mode and delay can be used in tandum:

           ``listen = :25 mode=bounce delay=10``

           The delay can also be specified as a range:

           ``listen = :25 delay=5-10``

           Using a delay range will cause the server to choose a random value
           within that range per connection.

           Mode and delay can be defined for each address/port in a listen
           directive:

           ``listen = :25 mode=bounce, :::25 delay=10, :587 mode=random``
        """
        if addr in ("127.0.0.1", "0.0.0.0"):  # nosec
            addr = ""
        elif addr in ("::1",):
            addr = "::"
        listeners = self.listen + self.tls_listen
        for laddr, lport, __, lflags in listeners:
            if laddr == addr and lport == port:
                return lflags
        return {}

    def create_flags(self, parts):
        """
        Create a set of flags from a listener directive.

        :param list parts: Parts of the listener definition.
        :returns: Flags for a listener. Default: ``{}``.
        :rtype: :py:obj:`dict`
        """
        flags = {}
        for part in parts:
            if part.count("=") == 1:
                flag, value = part.split("=")
                flag, value = flag.strip(), value.strip()
                if flag in ("mode", "delay"):
                    if flag == "mode":
                        flags.update(self._flag_mode(flag, value))
                    elif flag == "delay":
                        flags.update(self._flag_delay(flag, value))
        return flags

    def _flag_mode(self, flag, value):
        """
        Create a flag for the mode directive.

        :param str flag: The flag name.
        :param str value: The value of the flag.
        :returns: Mode flag for a listener.
        :rtype: :py:obj:`dict`
        :raises ConfigException: If an invalid mode is provided.
        """
        if value in ("accept", "bounce", "random"):
            return {flag: value}
        else:
            raise ConfigException(
                f"'{value}' is not a valid mode. Valid options "
                "are: 'accept', 'bounce' and 'random'.",
            )

    def _flag_delay(self, flag, value):
        """
        Create a delay flag, delay can be an int or a range.

        :param str flag: The flag name.
        :param str value: The value of the flag.
        :returns: Delay flag for a listener.
        :rtype: :py:obj:`dict`
        :raises ConfigException: If an invalid delay is provided.
        """
        if value.count("-") == 0:
            if value.isdigit() and int(value) < 60:
                return {flag: value}
            else:
                raise ConfigException(
                    f"'{value}' is not a valid delay. Delay is "
                    "in seconds and must be below 60.",
                )
        if value.count("-") == 1:
            start, end = value.split("-")
            start, end = start.strip(), end.strip()
            if (
                start.isdigit()
                and end.isdigit()
                and int(start) < 60
                and int(end) < 60
                and int(end) > int(start)
            ):
                return {flag: (start, end)}
        raise ConfigException(
            f"'{value}' is not a valid delay value. It must be "
            "either a single value or a range i.e. 5-10 and "
            "must be less than 60.",
        )

    def test(self):
        r"""
        Test configuration validity.

        :returns: The configuration object.
        :rtype: :class:`Config`

        .. note::

           Uses the magic of :py:func:`inspect.getmembers` to introspect
           methods beginning with \'test\_\' and calling them.
        """
        members = inspect.getmembers(self, predicate=inspect.ismethod)
        for name, _ in members:
            if name.startswith("test_"):
                getattr(self, name)()
        return self

    def test_workers(self):
        """
        Validate the number of workers.

        :raises ConfigException: If an invalid number of workers is provided.

        .. note::

           Cannot have more workers than number of processors or cores.
        """
        cpus = multiprocessing.cpu_count()
        if self.workers > cpus:
            msg = (
                "Cannot have more workers than number of processors or "
                f"cores. {self.workers} workers > {cpus} processors/cores."
            )
            raise ConfigException(msg)

    def test_ipv6_support(self):
        """
        If an IPv6 listener is configured, confirm IPv6 is supported.

        :raises ConfigException: If IPv6 is configured but is not supported by
                                 the operating system.
        """
        for address, __, family, ___ in self.listen:
            if ":" in address:
                if not socket.has_ipv6 and family == socket.AF_UNSPEC:
                    msg = (
                        "An IPv6 listener is configured but IPv6 is not "
                        "available on this platform."
                    )
                    raise ConfigException(msg)

    def test_tls_ipv6_support(self):
        """
        If an IPv6 listener is configured, confirm IPv6 is supported.

        :raises ConfigException: If IPv6 is configured but is not supported by
                                 the operating system.
        """
        for address, __, family, ___ in self.tls_listen:
            if ":" in address:
                if not socket.has_ipv6 and family == socket.AF_UNSPEC:
                    msg = (
                        "An IPv6 listener is configured but IPv6 is not "
                        "available on this platform."
                    )
                    raise ConfigException(msg)

    def test_same_listeners(self):
        """
        Test that multiple listeners are not configured on the same port.

        :raises ConfigException: When multiple listeners are configured on the
                                 same port.

        .. note::

           IPv4 and IPv6 addresses are different sockets so they can listen on
           the same port because they have different addresses.
        """
        if len(self.listen) == 0 and len(self.tls_listen) == 0:
            return
        listen, tls_listen = [], []
        for llisten in self.listen:
            addr, port, family, flags = llisten
            listen.append((addr, port, family))
        for llisten in self.tls_listen:
            addr, port, family, flags = llisten
            tls_listen.append((addr, port, family))
        if set(listen).intersection(tls_listen):
            msg = (
                "Cannot have multiple listeners on the same address and "
                "port."
            )
            raise ConfigException(msg)

    def test_no_listeners(self):
        """
        Test that at least one listener is configured.

        :raises ConfigException: When no listeners are configured.
        """
        if not len(self.listen) > 0 and not len(self.tls_listen) > 0:
            msg = "You need to define at least one listener."
            raise ConfigException(msg)

    def _min_max_port(self, port):
        """
        Minimum and maximum allowed port.

        :param int port: The port to test for validity.
        :raises ConfigException: When port is outside of the allowed range.

        .. note::

           On Linux the minimum is 1 and maximum is 65535.
        """
        min_port, max_port = 1, 65535
        if port < min_port:
            msg = (
                f"Port number {port} is not usable because it is less than "
                f"{min_port} which is the lowest available port."
            )
            raise ConfigException(msg)
        if port > max_port:
            msg = (
                f"Port number {port} is not usable because it is less than "
                f"{max_port} which is the highest available port."
            )
            raise ConfigException(msg)

    def test_port(self):
        """
        Validate port number.

        :raises ConfigException: When a port is configured that we have no
                                 permissions for.
        """
        for host, port, family, __ in self.listen:
            self._port_permissions(host, port, family)

    def _port_permissions(self, address, port, family):
        """
        Validate that we have permission to use the port and it's not in use.

        :param str address: The address to use.
        :param int port: The port to use.
        :param family: The type of socket to use.
        :type family: :py:obj:`socket.AF_INET` or :py:obj:`socket.AF_INET6`
        :raises ConfigException: When a port is supplied that we have no
                                 permissions for.
        """
        self._min_max_port(port)
        if os.getuid() != 0 and port < 1024:
            msg = f"You do not have permission to use port {port}"
            raise ConfigException(msg)
        sock = socket.socket(family, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except (AttributeError, OSError):
            pass
        if family == socket.AF_INET6:
            sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
        try:
            sock.bind((address, port))
        except OSError as err:
            errmsg = err.strerror
            msg = f"Could not use port {port}, {errmsg}"
            raise ConfigException(msg)
        finally:
            sock.close()
            del sock

    def test_user(self):
        """
        Validate user exists in UNIX password database.

        :raises ConfigException: When the user cannot be accessed on the
                                 operating system.

        .. note::

           Defaults to :py:func:`getpass.getuser` if no user is specified.
        """
        try:
            pwd.getpwnam(self.user)
        except KeyError:
            msg = f"{self._user} is not a valid user."
            raise ConfigException(msg)

    def test_group(self):
        """
        Validate group exists in UNIX group database.

        :raises ConfigException: When the group cannot be accessed on the
                                 operating system.

        .. note::

           Defaults to :py:attr:`grp.getgrgid.gr_name` if no group is
           specified.
        """
        try:
            grp.getgrnam(self.group)
        except KeyError:
            msg = f"{self._group} is a not a valid group."
            raise ConfigException(msg)

    def test_timeout(self):
        """
        Validate timeout - only allow a valid integer value in seconds.

        :raises ConfigException: When the timeout is not a number or is above
                                 the maximum allowed value of 180.
        """
        try:
            __ = self.timeout  # NOQA
        except ValueError:
            msg = f"{self._timeout} is not a valid number of seconds."
            raise ConfigException(msg)
        if self.timeout and self.timeout > 180:
            msg = (
                "Timeout must be at least 180 seconds or less for security "
                "(denial of service)."
            )
            raise ConfigException(msg)

    def test_tls_port(self):
        """
        Validate TLS port number.

        :raises ConfigException: When a port is configured that we have no
                                 permissions for.
        """
        if len(self.tls_listen) == 0:
            return
        for host, port, af, __ in self.tls_listen:
            self._port_permissions(host, port, af)

    def test_tls_settings(self):
        """
        Validate TLS configuration.

        :raises ConfigException: When the TLS configuration is invalid.

        .. note::

           Verifies if you provide all TLS settings, not just some.
        """
        port = True
        if not len(self.tls_listen) > 0:
            port = False
        cert = os.access(self.tls_cert, os.R_OK) if self.tls_cert else False
        key = os.access(self.tls_key, os.R_OK) if self.tls_key else False
        if (port, cert, key) == (False, False, False):
            return
        if not all((port, cert, key)):
            msg = (
                "To use TLS you must supply a port, certificate file "
                "and key file."
            )
            raise ConfigException(msg)

    def test_tls_dhparams(self):
        """
        Validate Diffie Hellman ephemeral parameters.

        :raises ConfigException: When the dhparams file is invalid.

        .. note::

           Verifies Diffie Hellman ephemeral parameters are readable.
        """
        if self.tls_dhparams and not os.access(self.tls_dhparams, os.R_OK):
            msg = (
                "To use Diffie Hellman ephemeral params you must supply a "
                "valid dhparams file."
            )
            raise ConfigException(msg)

    def test_delay(self):
        """
        Validate the delay period.

        :raises ConfigException: When the delay is not a number or is above
                                 the maximum allowed value of 60.

        .. note::

           Delay must be lower than the timeout.
        """
        if self.delay and self.delay >= self.timeout:
            msg = "Delay must be lower than timeout."
            raise ConfigException(msg)
        if self.delay and self.delay > 60:
            msg = (
                "Delay must be 60 seconds or less for security (denial of "
                "service)."
            )
            raise ConfigException(msg)

    def test_mode(self):
        """
        Validate the response mode.

        :raise ConfigException: When an invalid mode is configured.

        .. note::

           Valid options are: 'accept', 'bounce' and 'random'.
        """
        if self.mode not in ("accept", "bounce", "random"):
            msg = "Mode must be accept, bounce or random."
            raise ConfigException(msg)

    def test_max_message_size(self):
        """
        Validate max_message size is an integer.

        :raise ConfigException: When the maximum message size is not a number.
        """
        try:
            __ = self.max_message_size  # NOQA
        except ValueError:
            size = self._max_message_size
            msg = f"{size} is not a valid number of bytes."
            raise ConfigException(msg)

    def test_pidfile(self):
        """
        Validate that the pidfile can be written to.

        :raises ConfigException: When the pidfile is invalid.
        """
        if not self.pidfile:
            return
        pidfile = pathlib.Path(self.pidfile)
        if not os.access(pidfile.parent, os.W_OK):
            msg = "You do not have permission to write to the pidfile."
            raise ConfigException(msg)

    def test_dynamic_switch(self):
        """
        Validate that the dynamic_switch value is correct.

        :raises ConfigException: When the dynamic switch value is invalid.
        """
        if self._dynamic_switch is None:
            return
        if self._dynamic_switch not in (True, False):
            msg = "Allowed dynamic_switch values are true and false."
            raise ConfigException(msg)



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides control functionality, including socket wrappers."""

try:  # pragma: no cover
    import ssl
except ImportError:  # pragma: no cover
    ssl = None

import grp
import logging
import os
import pwd
import socket

from .config import Config
from .exceptions import BlackholeRuntimeException


__all__ = ("pid_permissions", "server", "setgid", "setuid")
"""Tuple all the things."""


logger = logging.getLogger("blackhole.control")
ciphers = [
    "ECDHE-ECDSA-AES256-GCM-SHA384",
    "ECDHE-RSA-AES256-GCM-SHA384",
    "ECDHE-ECDSA-CHACHA20-POLY1305",
    "ECDHE-RSA-CHACHA20-POLY1305",
    "ECDHE-ECDSA-AES128-GCM-SHA256",
    "ECDHE-RSA-AES128-GCM-SHA256",
    "ECDHE-ECDSA-AES256-SHA384",
    "ECDHE-RSA-AES256-SHA384",
    "ECDHE-ECDSA-AES128-SHA256",
    "ECDHE-RSA-AES128-SHA256",
]
"""Strong default TLS ciphers."""


def _context(use_tls=False):
    """
    Create a TLS context using the certificate, key and dhparams file.

    :param bool use_tls: Whether to create a TLS context or not.
                         Default: ``False``.
    :returns: A TLS context or ``None``.
    :rtype: :py:class:`ssl.SSLContext` or :py:obj:`None`.

    .. note::

       Created with:

       - :py:obj:`ssl.OP_NO_SSLv2`
       - :py:obj:`ssl.OP_NO_SSLv3`
       - :py:obj:`ssl.OP_NO_COMPRESSION`
       - :py:obj:`ssl.OP_CIPHER_SERVER_PREFERENCE`

       Also responsible for loading Diffie Hellman ephemeral parameters if
       they're provided -- :py:func:`ssl.SSLContext.load_dh_params`

       If the ``-ls`` or ``--less-secure`` option is provided,
       :py:obj:`ssl.OP_SINGLE_DH_USE` and :py:obj:`ssl.OP_SINGLE_ECDH_USE`
       will be omitted from the context. --
       https://kura.gg/blackhole/configuration.html#command-line-options
       -- added in :ref:`2.0.13`
    """
    if use_tls is False:
        return None
    config = Config()
    ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ctx.load_cert_chain(config.tls_cert, config.tls_key)
    ctx.options |= ssl.OP_NO_SSLv2
    ctx.options |= ssl.OP_NO_SSLv3
    ctx.options |= ssl.OP_NO_COMPRESSION
    ctx.options |= ssl.OP_CIPHER_SERVER_PREFERENCE
    if not config.args.less_secure:
        ctx.options |= ssl.OP_SINGLE_DH_USE
        ctx.options |= ssl.OP_SINGLE_ECDH_USE
    ctx.set_ciphers(":".join(ciphers))
    if config.tls_dhparams:
        ctx.load_dh_params(config.tls_dhparams)
    return ctx


def _socket(addr, port, family):
    """
    Create a socket, bind and listen.

    :param str addr: The address to use.
    :param int port: The port to use.
    :param family: The type of socket to use.
    :type family: :py:obj:`socket.AF_INET` or :py:obj:`socket.AF_INET6`.
    :returns: Bound socket.
    :rtype: :py:func:`socket.socket`
    :raises BlackholeRuntimeException: When a socket cannot be bound.
    """
    sock = socket.socket(family, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    try:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
    except (AttributeError, OSError):
        pass
    if family == socket.AF_INET6:
        sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_V6ONLY, 1)
    try:
        sock.bind((addr, port))
    except OSError:
        msg = f"Cannot bind to {addr}:{port}."
        logger.critical(msg)
        sock.close()
        raise BlackholeRuntimeException(msg)
    os.set_inheritable(sock.fileno(), True)
    sock.listen(1024)
    sock.setblocking(False)
    return sock


def server(addr, port, family, use_tls=False):
    """
    Socket and possibly a TLS context.

    Create an instance of :py:func:`socket.socket`, bind it and return a
    dictionary containing the socket object and a TLS context if configured.

    :param str addr: The address to use.
    :param int port: The port to use.
    :param family: The type of socket to use.
    :type family: :py:obj:`socket.AF_INET` or :py:obj:`socket.AF_INET6`.
    :param bool use_tls: Whether to create a TLS context or not.
                         Default: ``False``.
    :returns: Bound socket, a TLS context if configured.
    :rtype: :py:obj:`dict`
    """
    sock = _socket(addr, port, family)
    ctx = _context(use_tls=use_tls)
    return {"sock": sock, "ssl": ctx}


def pid_permissions():
    """
    Change the pid file ownership.

    Called before :func:`blackhole.control.setgid` and
    :func:`blackhole.control.setuid` are called to stop
    :class:`blackhole.daemon.Daemon` losing permissions to modify the pid file.

    :raises SystemExit: With exit code :py:obj:`os.EX_USAGE` when a permissions
                        error occurs.
    """
    config = Config()
    try:
        user = pwd.getpwnam(config.user)
        group = grp.getgrnam(config.group)
        os.chown(config.pidfile, user.pw_uid, group.gr_gid)
    except (KeyError, PermissionError):
        logger.error("Unable to change pidfile ownership permissions.")
        raise SystemExit(os.EX_USAGE)


def setgid():
    """
    Change group.

    Change to a less privileged group. Unless you're using it incorrectly --
    in which case, don't use it.

    :raises SystemExit: Exit code :py:obj:`os.EX_USAGE` when a configuration
                        error occurs or :py:obj:`os.EX_NOPERM` when a
                        permission error occurs.

    .. note::

       MUST be called BEFORE setuid, not after.
    """
    config = Config()
    try:
        gid = grp.getgrnam(config.group).gr_gid
        os.setgid(gid)
    except KeyError:
        logger.error(f"Group '{config.group}' does not exist.")
        raise SystemExit(os.EX_USAGE)
    except PermissionError:
        logger.error(
            f"You do not have permission to switch to group '{config.group}'.",
        )
        raise SystemExit(os.EX_NOPERM)


def setuid():
    """
    Change user.

    Change to a less privileged user.Unless you're using it incorrectly --
    inwhich case, don't use it.

    :raises SystemExit: Exit code :py:obj:`os.EX_USAGE` when a configuration
                        error occurs or :py:obj:`os.EX_NOPERM` when a
                        permission error occurs.

    .. note::

       MUST be called AFTER setgid, not before.
    """
    config = Config()
    try:
        uid = pwd.getpwnam(config.user).pw_uid
        os.setuid(uid)
    except KeyError:
        logger.error(f"User '{config.user}' does not exist.")
        raise SystemExit(os.EX_USAGE)
    except PermissionError:
        logger.error(
            f"You do not have permission to switch to user '{config.user}'.",
        )
        raise SystemExit(os.EX_NOPERM)



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Configure logging."""


import logging
from logging.config import dictConfig


__all__ = ("configure_logs",)
"""Tuple all the things."""


DEBUG_FORMAT = (
    "[%(asctime)s] [%(levelname)s] blackhole.%(module)s: %(message)s"
)

LOG_CONFIG = {
    "formatters": {
        "console": {"format": "%(message)s"},
        "debug": {"format": DEBUG_FORMAT},
    },
    "handlers": {},
    "loggers": {"blackhole": {"handlers": [], "level": logging.INFO}},
    "version": 1,
}

DEBUG_HANDLER = {
    "class": "logging.StreamHandler",
    "formatter": "debug",
    "level": logging.DEBUG,
}

DEFAULT_HANDLER = {
    "class": "logging.StreamHandler",
    "formatter": "console",
    "level": logging.INFO,
}


def configure_logs(args):
    """
    Configure the logging module.

    :param argparse.Namespace args: Parameters parsed from :py:mod:`argparse`.
    """
    logger_handlers = LOG_CONFIG["loggers"]["blackhole"]["handlers"]
    if args.debug:
        LOG_CONFIG["loggers"]["blackhole"]["level"] = logging.DEBUG
        LOG_CONFIG["handlers"]["default_handler"] = DEBUG_HANDLER
        logger_handlers.append("default_handler")
    elif args.quiet:
        LOG_CONFIG["loggers"]["blackhole"]["level"] = logging.ERROR
        LOG_CONFIG["handlers"]["default_handler"] = DEFAULT_HANDLER
        LOG_CONFIG["handlers"]["default_handler"]["level"] = logging.ERROR
        logger_handlers.append("default_handler")
    else:
        LOG_CONFIG["loggers"]["blackhole"]["level"] = logging.INFO
        LOG_CONFIG["handlers"]["default_handler"] = DEFAULT_HANDLER
        logger_handlers.append("default_handler")
    dictConfig(LOG_CONFIG)



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides utility functionality."""

import codecs
import os
import random
import socket
import time


__all__ = ("blackhole_config_help", "mailname", "message_id", "get_version")


class Singleton(type):
    """Singleton."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Override the __call__ method."""
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(
                *args,
                **kwargs,
            )
        return cls._instances[cls]


def mailname(mailname_file="/etc/mailname"):
    """
    Fully qualified domain name for HELO and EHLO.

    :param str mailname_file: A path to the mailname file. Default:
                              ``/etc/mailname``.
    :returns: A domain name.
    :rtype: :py:obj:`str`

    .. note::

       Prefers content of `mailname_file`, falls back on
       :py:func:`socket.getfqdn` if `mailname_file` does not exist or cannot be
       opened for reading.
    """
    if os.access(mailname_file, os.R_OK):
        with codecs.open(mailname_file, encoding="utf-8") as _mailname_file:
            mailname_content = _mailname_file.readlines()
            if len(mailname_content) == 0:
                return socket.getfqdn()
            nmailname_content = mailname_content[0].strip()
            if nmailname_content != "":
                return str(nmailname_content)
    return socket.getfqdn()


def message_id(domain):
    """
    Return a string suitable for RFC 2822 compliant Message-ID.

    :param str domain: A fully qualified domain.
    :returns: An RFC 2822 compliant Message-ID.
    :rtype: :py:obj:`str`
    """
    timeval = int(time.monotonic() * 100)
    pid = os.getpid()
    randint = random.getrandbits(64)
    return f"<{timeval}.{pid}.{randint}@{domain}>"


def get_version():
    """
    Extract the __version__ from a file without importing it.

    :return: The version that was extracted.
    :rtype: :py:obj:`str`
    :raises AssertionError: When a version cannot be determined.
    """
    path = os.path.dirname(os.path.abspath(__file__))
    filepath = os.path.join(path, "__init__.py")
    if not os.access(filepath, os.R_OK):
        raise OSError("Cannot open __init__.py file for reading")
    with codecs.open(filepath, encoding="utf-8") as fp:
        for line in fp:
            if line.startswith("__version__"):
                try:
                    _, vers = line.split("=")
                except ValueError:
                    msg = "Cannot extract version from __version__"
                    raise AssertionError(msg)
                version = vers.strip().replace('"', "").replace("'", "")
                try:
                    major, minor, patch = version.split(".")
                    digits = (
                        major.isdigit(),
                        minor.isdigit(),
                        patch.isdigit(),
                    )
                    if not all(digits):
                        msg = f"{version} is not a valid version number"
                        raise AssertionError(msg)
                except ValueError:
                    msg = f"{version} is not a valid version number"
                    raise AssertionError(msg)
                return version
    raise AssertionError("No __version__ assignment found")


class Formatter:
    """Terminal output formatter."""

    bold = "\033[1m"
    under = "\033[4m"
    reset = "\033[0m"


formatting = Formatter()

# fmt: off
blackhole_config_help = '''BLACKHOLE_CONFIG(1)

{f.bold}NAME{f.reset}
       blackhole_config - the config file format for the Blackhole MTA

{f.bold}DESCRIPTION{f.reset}
       This manual page documents the {f.bold}Blackhole{f.reset} configuration file format and
       options.

{f.bold}OPTIONS{f.reset}
    These are all available options for the configuration file, their default
    values and information on what the options actually do.

    The file format is a simple {f.under}attribute = value{f.reset} style, an example is shown
    below.

        # This is a comment.
        listen = :25  # This is an inline comment.
        user = kura
        group = kura

    {f.bold}listen{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}listen{f.reset} = {f.under}[address]:port [mode=MODE] [delay=DELAY]{f.reset}

        {f.bold}Default{f.reset}
            127.0.0.1:25,  127.0.0.1:587, :::25, :::587
            IPv6 listeners are only enabled if IPv6 is supported.

        {f.bold}Optional{f.reset}
            {f.under}mode={f.reset} and {f.under}delay={f.reset} -- allows setting a response mode and delay per
            listener.

        The {f.under}mode={f.reset} and {f.under}delay={f.reset} flags allow specific ports to act in different ways.
        i.e. you could accept all mail on 10.0.0.1:25 and bounce it all on
        10.0.0.2:25, as below.

            listen = 10.0.0.1:25 mode=accept, 10.0.0.2:25 mode=bounce

        The {f.under}mode={f.reset} and {f.under}delay={f.reset} flags may also be specified together, as required.

            listen = 10.0.0.1:25 mode=accept delay=5, 10.0.0.2:25 mode=bounce delay=10

        The flags accept the same options as {f.under}dynamic-switches{f.reset}, including setting
        a delay range.

                                            ----

    {f.bold}tls_listen{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}tls_listen{f.reset} = {f.under}[address]:port [mode=MODE] [delay=DELAY]{f.reset}

        {f.bold}Default{f.reset}
            None

        {f.bold}Optional{f.reset}
            {f.under}mode={f.reset} and {f.under}delay={f.reset} -- allows setting a response mode and delay per
            listener.


        :465 is equivalent to listening on port 465 on all IPv4 addresses and
        :::465 is equivalent to listening on port 465 on all IPv6 addresses.

        Multiple addresses and ports can be listed on a single line.

            tls_listen = 10.0.0.1:465, 10.0.0.2:465, :465, :::465

        The {f.under}mode={f.reset} and {f.under}delay={f.reset} flags allow specific ports to act in different ways.
        i.e. you could accept all mail on 10.0.0.1:465 and bounce it all on
        10.0.0.2:465, as below.

            tls_listen = 10.0.0.1:465 mode=accept, 10.0.0.2:465 mode=bounce

        The {f.under}mode={f.reset} and {f.under}delay={f.reset} flags may also be specified together, as required.

            tls_listen = 10.0.0.1:465 mode=accept delay=5, 10.0.0.2:465 mode=bounce delay=10

        The flags accept the same options as {f.under}dynamic-switches{f.reset}, including setting
        a delay range.

                                            ----

    {f.bold}user{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}user{f.reset} = {f.under}user{f.reset}

        {f.bold}Default{f.reset}
            The current Linux user

        Blackhole will set it's process owner to the value provided with this
        option. Ports below 1024 require sudo or root privileges, this option
        is available so that the process can be started, listen on privileged
        ports and then give up those privileges.

                                            ----

    {f.bold}group{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}group{f.reset} = {f.under}group{f.reset}

        {f.bold}Default{f.reset}
            The primary group of the current Linux user

        Blackhole will set it's process group to the value provided with this
        option.

                                            ----

    {f.bold}pidfile{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}pidfile{f.reset} = {f.under}/path/to/file.pid{f.reset}

        {f.bold}Default{f.reset}
            /tmp/blackhole.pid

        Blackhole will write it's Process ID to this file, allowing you to
        easily track the process and send signals to it.

                                            ----

    {f.bold}timeout{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}timeout{f.reset} = {f.under}seconds{f.reset}

        {f.bold}Default{f.reset}
            60 -- Maximum value of 180 seconds.

        This is the amount of time to wait for a client to send data. Once the
        timeout value has been reached with no data being sent by the client,
        the connection will be terminated and a 421 Timeout message will be
        sent to the client.

        Helps mitigate DoS risks.

                                            ----

    {f.bold}tls_cert{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}tls_cert{f.reset} = {f.under}/path/to/certificate.pem{f.reset}

        {f.bold}Default{f.reset}
            None

        The certificate file in x509 format for wrapping a connection in
        SSL/TLS.

                                            ----

    {f.bold}tls_key{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}tls_key{f.reset} = {f.under}/path/to/private.key{f.reset}

        {f.bold}Default{f.reset}
            None

        The private key of the tls_cert.

                                            ----

    {f.bold}tls_dhparams{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}tls_dhparams{f.reset} = {f.under}/path/to/dhparams.pem{f.reset}

        {f.bold}Default{f.reset}
            None

        File containing Diffie Hellman ephemeral parameters for ECDH ciphers.

                                            ----

    {f.bold}delay{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}delay{f.reset} = {f.under}seconds{f.reset}

        {f.bold}Default{f.reset}
            0 -- Maximum value of 60 seconds.

        Time to delay before returning a response to a completed DATA command.
        You can use this to delay testing or simulate lag.

                                            ----

    {f.bold}mode{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}mode{f.reset} = {f.under}accept | bounce | random{f.reset}

        {f.bold}Default{f.reset}
            accept

                                            ----

    {f.bold}max_message_size{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}max_message_size{f.reset} = {f.under}bytes{f.reset}

        {f.bold}Default{f.reset}
            512000 Bytes (512 KB)

        The maximum message size for a message. This includes headers and helps
        mitigate a DoS risk.

                                            ----

    {f.bold}dynamic_switch{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}dynamic_switch{f.reset} = {f.under}true | false{f.reset}

        {f.bold}Default{f.reset}
            true

        The dynamic switch option allows you to enable or disable parsing of
        dynamic switches from email headers.

                                            ----

    {f.bold}workers{f.reset}
        {f.bold}Syntax{f.reset}
            {f.bold}workers{f.reset} = {f.under}number{f.reset}

        {f.bold}Default{f.reset}
            1

        The workers option allows you to define how many worker processes to
        spawn to handle incoming mail. The absolute minimum is actually 2. Even
        by setting the workers value to 1, a supervisor process will always
        exist meaning that you would have 1 worker and a supervisor.
'''.format(f=formatting)  # noqa
# fmt: on



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides functionality to create the supervisor process."""


import asyncio
import logging
import os
import signal

from .config import Config
from .control import server
from .exceptions import BlackholeRuntimeException
from .utils import Singleton
from .worker import Worker


try:
    import setproctitle
except ImportError:
    setproctitle = None


__all__ = ("Supervisor",)
"""Tuple all the things."""


logger = logging.getLogger("blackhole.supervisor")


class Supervisor(metaclass=Singleton):
    """
    The supervisor process.

    Responsible for monitoring and controlling child processes via an
    internal map of workers and the children they manage.
    """

    def __init__(self, loop=None):
        """
        Initialise the supervisor.

        Loads the configuration and event loop.

        :param loop: The event loop to use.
        :type loop: :py:class:`syncio.unix_events._UnixSelectorEventLoop` or
                    :py:obj:`None` to get the current event loop using
                    :py:func:`asyncio.get_event_loop`.
        :raises BlackholeRuntimeException: When an error occurs generating
                                           servers.
        """
        logger.debug("Initiating the supervisor")
        self.config = Config()
        self.loop = loop if loop is not None else asyncio.get_event_loop()
        self.socks = []
        self.workers = []
        if setproctitle:
            setproctitle.setproctitle("blackhole: master")
        try:
            self.generate_servers()
        except BlackholeRuntimeException:
            self.close_socks()
            raise BlackholeRuntimeException()

    def generate_servers(self):
        """Spawn all of the required sockets and TLS contexts."""
        logger.debug("Attaching sockets to the supervisor")
        self.create_socket(self.config.listen)
        tls_conf = (self.config.tls_cert, self.config.tls_key)
        if len(self.config.tls_listen) > 0 and all(tls_conf):
            self.create_socket(self.config.tls_listen, use_tls=True)

    def create_socket(self, listeners, use_tls=False):
        """Create supervisor socket."""
        _tls = ""
        if use_tls:
            _tls = " (TLS)"
        for host, port, family, flags in listeners:
            aserver = server(host, port, family, use_tls=use_tls)
            self.socks.append(aserver)
            logger.debug(f"Attaching {host}:{port}{_tls} with flags {flags}")

    def run(self):
        """
        Start all workers and their children.

        Attaches signals and runs the event loop.
        """
        self.start_workers()
        signal.signal(signal.SIGTERM, self.stop)
        signal.signal(signal.SIGINT, self.stop)
        self.loop.run_forever()

    def start_workers(self):
        """Start each worker and it's child process."""
        logger.debug("Starting workers")
        for idx in range(self.config.workers):
            num = f"{idx + 1}"
            logger.debug(f"Creating worker: {num}")
            self.workers.append(Worker(num, self.socks, self.loop))

    def stop_workers(self):
        """Stop the workers and their respective child process."""
        logger.debug("Stopping workers")
        worker_num = 1
        for worker in self.workers:
            logger.debug(f"Stopping worker: {worker_num}")
            worker.stop()
            worker_num += 1

    def close_socks(self):
        """Close all opened sockets."""
        for sock in self.socks:
            sock["sock"].close()

    def stop(self, *args, **kwargs):
        """
        Terminate all of the workers.

        Generally should be called by a signal, nothing else.

        :raise SystemExit: With code :py:obj:`os.EX_OK`.
        """
        self.stop_workers()
        self.close_socks()
        logger.debug("Stopping supervisor")
        try:
            self.loop.stop()
        except RuntimeError:
            pass
        raise SystemExit(os.EX_OK)



# -*- coding: utf-8 -*-

# (The MIT License)
#
# Copyright (c) 2013-2021 Kura
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides functionality to spawn and control child processes."""


import asyncio
import logging
import os
import signal

from . import protocols
from .smtp import Smtp
from .streams import StreamProtocol


__all__ = ("Child",)
"""Tuple all the things."""


logger = logging.getLogger("blackhole.child")


class Child:
    """
    A child process.

    Each child process maintains a list of the internal
    :py:class:`asyncio.Server` instances it utilises. Each child also
    maintains a list of all connections being managed by the child.
    """

    _started = False
    servers = []
    """List of :py:class:`asyncio.Server` instances."""

    clients = []
    """List of clients connected to this process."""

    def __init__(self, up_read, down_write, socks, idx):
        """
        Initialise a child process.

        :param int up_read: A file descriptor for reading.
        :param int down_write: A file descriptor for writing.
        :param list socks: A list of sockets.
        """
        self.up_read = up_read
        self.down_write = down_write
        self.socks = socks
        self.idx = idx

    def start(self):
        """Start the child process."""
        logger.debug(f"Starting child {self.idx}")
        self._started = True
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)
        signal.signal(signal.SIGTERM, self.stop)
        self.heartbeat_task = asyncio.Task(self.heartbeat())
        self.loop.run_forever()
        self.stop()
        os._exit(os.EX_OK)

    async def _start(self):
        """Create an asyncio server for each socket."""
        for sock in self.socks:
            server = await self.loop.create_server(
                lambda: Smtp(self.clients),
                **sock,
            )
            self.servers.append(server)

    def stop(self, *args, **kwargs):
        """
        Stop the child process.

        Mark the process as being stopped, closes each client connected via
        this child, cancels internal communication with the supervisor and
        finally stops the process and exits.
        """
        self._started = False
        for _ in range(len(self.clients)):
            client = self.clients.pop()
            client.close()
        for _ in range(len(self.servers)):
            server = self.servers.pop()
            server.close()
        self.heartbeat_task.cancel()
        self.server_task.cancel()
        for task in asyncio.all_tasks(self.loop):
            task.cancel()
        self.loop.stop()
        self._started = False
        os._exit(os.EX_OK)

    async def heartbeat(self):
        """
        Handle heartbeat between a worker and child.

        If a child process stops communicating with it's worker, it will be
        killed, the worker managing it will also be removed and a new worker
        and child will be spawned.

        .. note::

           3 bytes are used in the communication channel.

           - b'x01' -- :const:`blackhole.protocols.PING`
           - b'x02' -- :const:`blackhole.protocols.PONG`

           These message values are defined in the :mod:`blackhole.protocols`
           schema. Documentation is available at --
           https://kura.gg/blackhole/api-protocols.html
        """
        read_fd = os.fdopen(self.up_read, "rb")
        r_trans, r_proto = await self.loop.connect_read_pipe(
            StreamProtocol,
            read_fd,
        )
        write_fd = os.fdopen(self.down_write, "wb")
        w_trans, w_proto = await self.loop.connect_write_pipe(
            StreamProtocol,
            write_fd,
        )
        reader = r_proto.reader
        writer = asyncio.StreamWriter(w_trans, w_proto, reader, self.loop)
        self.server_task = asyncio.Task(self._start())

        while self._started:
            try:
                msg = await reader.read(3)
            except:  # noqa
                break
            if msg == protocols.PING:
                logger.debug(
                    f"child.{self.idx}.heartbeat: Ping request received from "
                    "parent",
                )
                writer.write(protocols.PONG)
            await asyncio.sleep(5)
        r_trans.close()
        w_trans.close()
        self.stop()
